from dataikuapi.dss.ml import DSSPredictionMLTaskSettings

class DSSMLTaskSettingsExtended(DSSPredictionMLTaskSettings): 
    
    
    def foreach_feature(self, fn, only_of_type = None):
        """
        Applies a function to all features (except target)
        
        :param function fn: Function that takes 2 parameters: feature_name and feature_params and returns modified feature_params
        :param str only_of_type: if not None, only applies to feature of the given type. Can be one of ``CATEGORY``, ``NUMERIC``, ``TEXT`` or ``VECTOR``
        """
        import copy
        new_per_feature = {}
        for (k, v) in self.mltask_settings["preprocessing"]["per_feature"].items():
            if v["role"] == "TARGET":
                continue
            if only_of_type is not None and v["type"] != only_of_type:
                continue
            new_per_feature[k] = fn(k, copy.deepcopy(v))
        self.mltask_settings["preprocessing"]["per_feature"].update(new_per_feature)
    
    
    """FEATURE SETTING"""
    def get_mass_feature_preprocessing(self, criteria_dict = None):
        
        """        
        Gets the feature preprocessing params for all features whose settings satisfy the conditions in
        criteria_dict. 
        This returns a reference to the feature's settings, not a copy, so changes made to the returned 
        dict will be reflected when saving. 
        However, it is advised to use the :meth:`DSSMLTaskSettingsExtended.set_mass_feature_setting`.
        
        :param dict criteria_dict: Conditions to filter features. 
        
        :returns: A setting dict of features satisfying the conditions.
        :rtype: dict
        """
        
        # we begin with the full feature list
        new_feat_dict = self.get_raw()['preprocessing']['per_feature'].copy()
                    
        if criteria_dict is not None:
            
            for (feature_name, feature_setting) in self.get_raw()['preprocessing']['per_feature'].items():

                # check if the temp_feature_preprocessing_dict DOESNT satisfy the criteria dict, if that's 
                # the case, remove the feature from the final dict
                if not(criteria_dict.viewitems() < feature_setting.viewitems()): 
                    new_feat_dict.pop(feature_name)
                        
        return new_feat_dict     
        
    
    def set_mass_feature_setting(self, criteria_dict, new_setting_dict):
        
        """        
        For all features whose settings satisfy the criteria_dict, modify their settings using the new value in 
        new_setting_dict.
        
        :param dict criteria_dict: Conditions to filter features.
        :param dict new_setting_dict: The new settings.
        """        
        
        feature_setting_dict = self.get_raw()['preprocessing']['per_feature'] # dict of dict: {FEATURE_NAME:{settings_dict}, }
        
        for (feature_name, feature_setting) in feature_setting_dict.items(): 
            
            # check if the settings of the actual feature satisfies the criteria dict
            if criteria_dict.viewitems() < feature_setting.viewitems():
                if len(set(new_setting_dict.keys()) - set(feature_setting.keys())) > 0:
                    invalid_keys = list(set(new_setting_dict.keys()) - set(feature_setting.keys()))
                    raise ValueError('Invalid parameter in the setting dict: {}'.format(invalid_keys))
                
                feature_setting.update(new_setting_dict) # update the parameters

                
    def set_feature_setting(self, feature_name, new_setting_dict):
         
        """        
        For the single feature whose name is feature_name, modify its settings using the new value in 
        new_setting_dict.
        
        :param str feature_name: Name of the feature to modify.
        :param dict new_setting_dict: Dictionary of new settings.
        
        """
            
        feature_setting = self.get_feature_preprocessing(feature_name)
        
        if len(set(new_setting_dict.keys()) - set(feature_setting.keys())) > 0:

            invalid_keys = list(set(new_setting_dict.keys()) - set(feature_setting.keys()))
            raise ValueError('Invalid parameter in the setting dict: {}'.format(invalid_keys))

        feature_setting.update(new_setting_dict) # update the parameters  

            
     
    """ALGORITHM SETTING"""
    def set_algorithm_list_enabled(self, algorithm_list, enabled = True):

        """  
        Enable/disable all algorithms in the chosen list. 
        The rest of them will be on the other state.
        
        :param list algorithm_list: List of algorithms to enable/disable.
        :param boolean enabled: Enable these algorithms or not. 
        """
        
        # retrieve list of algorithm's names
        all_algorithm_list = self.get_mass_algorithm_setting().keys()
        
        # first disable/enable all algo
        for algorithm_name in all_algorithm_list:
            self.set_algorithm_enabled(algorithm_name, enabled = not enabled)  
        
        # we then enable the selected ones
        for algorithm_name in algorithm_list:
            self.set_algorithm_enabled(algorithm_name, enabled = enabled)        
  

    def set_algorithm_setting(self, algorithm_name, new_setting_dict):
        
        """        
        Update the settings of an algorithm.
        
        :param str algorithm_name: Name of the algorithm.
        :param dict new_setting_dict: New settings for the algorithm.
        """
        
        algorithm_setting_dict = self.get_algorithm_settings(algorithm_name = algorithm_name) 
        
        if len(set(new_setting_dict.keys()) - set(algorithm_setting_dict.keys())) > 0:
            
            invalid_keys = list(set(new_setting_dict.keys()) - set(algorithm_setting_dict.keys()))
            raise ValueError('Invalid parameter in the setting dict: {}'.format(invalid_keys))
        
        algorithm_setting_dict.update(new_setting_dict)
        

    def get_mass_algorithm_setting(self, enabled = None):
        
        """        
        Gets the training settings for all/enabled/disabled algorithms (depending on the ``enabled`` parameter). 
        This returns a reference to the algorithm's settings, not a copy, so changes made to the returned object 
        will be reflected when saving.
        However, as with :meth:`DSSMLTaskSettingsExtended.get_mass_feature_settings`, it is advised to use dedicated
        set function :meth:`DSSMLTaskSettingsExtended.set_algorithm_setting`.

        All algorithms have at least an "enabled" setting. Other settings are algorithm-dependent. You can print
        the returned object to learn more about the settings of each particular algorithm.
        
        :param boolean enabled: choose if we want to retrieve all/enabled/disabled algorithms.
        
        :return: A dict of the settings for all/enabled/disabled algorithms.
        :rtype: dict 
        """
        
        new_algorithm_setting_dict = self.get_raw()['modeling'].copy()
        
        if enabled is None: # return all algorithms 
            for (algo_name, algo_dict) in self.get_raw()['modeling'].items():
                # if it is not an algorithm, remove it from the final dict
                if not(isinstance(algo_dict, dict) and 'enabled' in algo_dict.keys()):
                    new_algorithm_setting_dict.pop(algo_name)
        
        else: # choose only enabled/disabled algorithms
            for (algo_name, algo_dict) in self.get_raw()['modeling'].items():
                if not(isinstance(algo_dict, dict) and algo_dict.get('enabled') == enabled): 
                    new_algorithm_setting_dict.pop(algo_name)
        

        return new_algorithm_setting_dict
    
    
    """GRID SEARCH SETTING"""
    def get_gridsearch_setting(self):
    
        """        
        Get the gridsearch setting. 
        This returns a reference to the gridsearch's settings, not a copy, so changes made to the returned 
        object will be reflected when saving.
        To modify these setting, it is advised to use :meth:`DSSMLTaskSettingsExtended.set_gridsearch_setting`.        
        
        :returns: a dict of performance metrics values
        :rtype: dict
        """
        
        return self.get_raw()['modeling']['gridSearchParams']

    
    def set_gridsearch_setting(self, gridsearch_setting_dict):

        """        
        Set the gridsearch setting.
        
        :param dict gridsearch_setting_dict: New setting for gridsearch.
        """
        
        self.get_raw()['modeling']['gridSearchParams'].update(gridsearch_setting_dict)